//------------------------------------------------------------
// contains files for reading and playing Music Studio Files
// Created Feb 16, 2011
// by Jim Patchell....the blind hacker
//------------------------------------------------------------

//#include "stdafx.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include "MSplay.h"

/*
	GLOBAL VARIABLES
*/

STACK *pRepeatStack;
int ticks;

/*
	timer globals
*/

int track_enables[4];

//------------------------
// Not really sure which
// duration goes with
// which note...(this
// may have been wrong
// to start with)
//------------------------
typedef struct {
	int Dur;
	char *pName;
} DUR;

static DUR DurTab[0x15] = {
	0,"NA",				//0
	4,"32nd Trip",		//1 
	0,"NA",				//2
	6,"ThirtySecond",	//3 thirty second
	8,"16th Trip",				//4 
	9,"Dotted TS",		//5 Dotted Thirty Second	
	12,"Sixteenth",		//6 sixteenth		
	16,"8th Trip",		//7 
	18,"Dotted Sixteenth",	//8	Dotted Sixteenth	
	24,"Eighth",		//9 eighth		
	32,	"1/4 Trip",			//10 
	36,	"Dotted Eighth",//11 Dotted Eighth		
	48,"Quarter",		//12 quarter
	64,"1/2 Trip",			//13
	72,"Dotted Quarter",//14 Dotted Quarter Note
	96,"Half",			//15 half
	128,"Whole Trip",	//16
	144,"Dptted Half",	//17 Dotted Halfe
	192,"Whole",		//18 whole	
	0,"NA",				//19
	288,"Dotted Whole"	//20 dotted whole
};

int dur[0x15] = {
	0,
	4,		//thirty second
	0,
	6,		//dotted thirty second
	8,		//sixteenth
	9,		
	12,		//dotted sixteenth		
	16,		//eighth
	18,		
	24,		//dotted eighth		
	32,		//quarter
	36,		
	48,		//dotted quarter
	64,		//half
	72,		
	96,		//dotted half
	128,	//whole
	144,	
	192,	//dotted whole
	0,
	288		
};
int key_sig_tab[128];
int octave[6] = {0,-24,-12,0,12,24};
int s_octave[6][7];

int	active_timer = 0;
int done_handshake = 0;
char midi_str[80];	/*	midi note buffer	*/
int midi_str_len;

char *NoteLUT[12] = {
	"C",
	"C#",
	"D",
	"D#",
	"E",
	"F",
	"F#",
	"G",
	"G#",
	"A",
	"A#",
	"B"
};

static char *KeySigTab[] = {
	"ERROR",
	"CMAJ",
	"GMAJ",
	"DMAJ",
	"AMAJ",
	"EMAJ",
	"BMAJ",
	"F#MAJ",
	"C#MAJ",
	"FMAJ",
	"BbMAJ",
	"EbMAJ",
	"AbMAJ",
	"DbMAJ",
	"GbMAJ",
	"CbMAJ"
};

static char *TimeSigLut[] = {
	"NA",
	"2/2",
	"3/2",
	"2/4",//3
	"3/4",
	"4/4",
	"5/4",
	"6/8",
	"NA"
};

int NoteToStr(char *s,int pitch)
{
	int oct;
	int note;

	oct = pitch / 12;
	note = pitch % 12;
	return sprintf(s,"%s%d",NoteLUT[note],oct);
}

unsigned short LittleEndian(unsigned short u)
{
	union {
		unsigned short uv;
		unsigned char cv[2];
	}i,o;

	i.uv = u;
	o.cv[0] = i.cv[1];
	o.cv[1] = i.cv[0];
	return o.uv;
}

void StartSong(MSSONG *pSong)
{
	int i,j;
	short unsigned int *wptr;
	short unsigned int u,mask;
	char *SNGdata;

	InitBuffer(pSong);
	SNGdata = (char *)&pSong->SongData[MS_OFFSET_SONG_DATA];
	pSong->Ticks = 100;
//	set_timer();
	pSong->Event = 0;

	wptr = (unsigned short int *)SNGdata;
	for(i=0;i<4;++i,++wptr)
	{
		u = *wptr;	/*	get track enable info	*/
//		printf("U=%04x\n",u & 0xffff);
		u = LittleEndian(u);
//		printf("U=%04x\n",u & 0xffff);
		mask = 1;
		if(i == 0)
		{
			for(j=0;j<15;++j)
			{
				pSong->trkinfo[j].EnableFlag = (u & mask)?1:0;
				mask <<= 1;	//shift mask
			}
		}
	}
	pSong->pCurSong = (char *)wptr;
}
int parse(MSSONG *pSong,SEVENT **ppSe)
{
	//----------------------------------------
	// This function is used to parse the song
	// data so that it can be either played or
	// what ever.
	//
	// This function will continue to parse
	// untill it has decoded a complete event
	// or in other words, all of the things
	// that happen at the same time
	//
	// parameters:
	//	pSong......pointer to song descriptor
	//
	// return value
	//	returns pointer to SEVENT chain.  Returns 0 when
	//	song is finished.
	//----------------------------------------

	int i;
	int t1,t2;
	char *midi= new char[80];	/*	buffer for midi data to play	*/
	int repeat,loop;
	int rv = 1;
	EVENT *pEv = NULL;
	SEVENT *pFirst = NULL;
	SEVENT *pLast = NULL;
	SEVENT *pSe;
	int key_sig,time_sig;
	int tempo;

	loop = 1;
	while(loop)
	{
		i = ((int)(*pSong->pCurSong++)) & 0x0ff;
//		printf("MS Token = %08x\n",i);
		switch(i)	/*	switch song data	*/
		{
			case REPEAT_STOP:	//basically the same thing as an End Event
//				printf("Repeat Stop\n");
				pSe = newSongEvent(SEVENT_OBJ_TYPE);
				pSe->d.obj.ObjType = SOBJ_TYPE_REPEATEND;
				if(!pFirst)
				{
					pFirst = pSe;
					pLast = pSe;
				}
				else
				{
					pLast->next = pSe;
					pLast = pSe;
				}
			case EVENT_END:
				++pSong->Event;
//				printf("Event End = %d\n",pSong->Event);
				*ppSe = pFirst;
				loop = 0;
				break;
			case KEY_SIGNATURE:
				*pSong->pCurSong++;	//skip 0x01?
				key_sig = ((int)(*pSong->pCurSong++)) & 0x0ff;
//				printf("Key Signiture = %d\n",key_sig);
				pSe = newSongEvent(SEVENT_OBJ_TYPE);
				pSe->d.obj.ObjType = SOBJ_TYPE_KEY_SIGNATURE;
				pSe->d.obj.Data = key_sig;
				if(!pFirst)
				{
					pFirst = pSe;
					pLast = pSe;
				}
				else
				{
					pLast->next = pSe;
					pLast = pSe;
				}

//				gen_key_tab(key_sig);
				break;
			case TEMPO:
				tempo = (int)(*pSong->pCurSong++) & 0xff;
//				printf("Tempo = %d\n",tempo);
				if((tempo = 12288 / (( 48 * tempo) / 60)) > 255)
					tempo = 0;
				pSe = newSongEvent(SEVENT_OBJ_TYPE);
				pSe->d.obj.ObjType = SOBJ_TYPE_TEMPO;
				pSe->d.obj.Data = tempo;
				if(!pFirst)
				{
					pFirst = pSe;
					pLast = pSe;
				}
				else
				{
					pLast->next = pSe;
					pLast = pSe;
				}
				break;
			case BAR:	//don't do much of anything
				pSong->MeasureBarCnt ++;
//				printf("Measure Bar");
				pSe = newSongEvent(SEVENT_OBJ_TYPE);
				pSe->d.obj.ObjType = SOBJ_TYPE_MEASUREBAR;
				if(!pFirst)
				{
					pFirst = pSe;
					pLast = pSe;
				}
				else
				{
					pLast->next = pSe;
					pLast = pSe;
				}
				break;
			case TIME_SIGNATURE:
				time_sig = (int)(*pSong->pCurSong++);
//				printf("Time Signiture = %d\n",time_sig);
				pSe = newSongEvent(SEVENT_OBJ_TYPE);
				pSe->d.obj.ObjType = SOBJ_TYPE_TIMESIGNATURE;
				pSe->d.obj.Data = time_sig;
				if(!pFirst)
				{
					pFirst = pSe;
					pLast = pSe;
				}
				else
				{
					pLast->next = pSe;
					pLast = pSe;
				}
				break;
			case LOUDNESS:
				pSong->pCurSong++;	/*	points to 0x01	*/
				pSe = newSongEvent(SEVENT_OBJ_TYPE);
				pSe->d.obj.ObjType = SOBJ_TYPE_LOUDNESS;
				pSe->d.obj.Data = *pSong->pCurSong;
				if(!pFirst)
				{
					pFirst = pSe;
					pLast = pSe;
				}
				else
				{
					pLast->next = pSe;
					pLast = pSe;
				}
//				printf("Loudness = %d\n",*pSong->pCurSong);
				pSong->pCurSong++;	/*	points to next item	*/
				break;
			case REPEAT_START:
				repeat = (int)*pSong->pCurSong++;
//				printf("Start Repeat = %d\n",repeat);
				pSe = newSongEvent(SEVENT_OBJ_TYPE);
				pSe->d.obj.ObjType = SOBJ_TYPE_REPEATSTART;
				pSe->d.obj.Data = repeat;
				if(!pFirst)
				{
					pFirst = pSe;
					pLast = pSe;
				}
				else
				{
					pLast->next = pSe;
					pLast = pSe;
				}
				break;
			case END:
//				printf("End of Song\n");
				pSe = newSongEvent(SEVENT_OBJ_TYPE);
				pSe->d.obj.ObjType = SOBJ_TYPE_END;
				pSe->d.obj.Data = repeat;
				if(!pFirst)
				{
					pFirst = pSe;
					pLast = pSe;
				}
				else
				{
					pLast->next = pSe;
					pLast = pSe;
				}
				loop = 0;
				*ppSe = pFirst;
				rv = 0;
				break;
			default:
				/*
					if we got down here, what was found must be a note
					so, we first get the track/type,duration,accidentals
					and then the note number
					If, the note starts a tie, treat it like any other
					note, if a note is at the end of a tie, then find
					the note in the time out buffer.  If the note is
					the beginning and end of a tie, still find it
					in the timeout buffer
				*/
				t1 = (int)*pSong->pCurSong++;	/*	duration	*/
				t2 = ((int)*pSong->pCurSong++) & 0x7f;	/*	note mask off MSB (upside down note)	*/
				pSe = newSongEvent(SEVENT_NOTE_TYPE);
				pSe->d.note.TieEnd = (i & MS_ENDTIE_FLAG)?1:0;
				pSe->d.note.TieBeg = (i & MS_BEGTIE_FLAG)?1:0;
				pSe->d.note.Accent = (t1 & MS_ACCENT_FLAG)?1:0;
				pSe->d.note.Rest = (i & MS_REST_FLAG)?1:0;
				pSe->d.note.Duration = t1 & MS_DURATION_MASK;
				pSe->d.note.Key = t1 & MS_KEY_MASK;
				pSe->d.note.Track = i & MS_TRACK_MASK;
				pSe->d.note.Note = t2;
				if(!pFirst)
				{
					pFirst = pSe;
					pLast = pSe;
				}
				else
				{
					pLast->next = pSe;
					pLast = pSe;
				}

				break;
		}	/*	end of switch i	*/
	}	/*	end while	*/
	return rv;
}

EVENT *search(int note,int chan,EVENT *pEv, int MaxEvents)
{
	//---------------------------------
	// search events for a particular
	// note on a particular channel
	//
	// parameters:
	//	note....note to search for
	//	chan....on this particular channel
	//	pEv.....pointer to an array of EVENT objects
	//	MaxEvents...maximum number of events in array
	//
	// return value
	//	Returns NULL if no event was found
	//	otherwise it returns a pointer to
	//	the desired object
	//---------------------------------

	int i,loop;
	EVENT *pE;

	for(i=0,loop = 1;(i<MaxEvents)&&loop;++i)
	{
		if((pEv[i].Note == note) && !pEv[i].RestFlag && (chan == pEv[i].Channel) )
		{
			loop = 0;
			pE = &pEv[i];
		}
	}
	return(pE);
}

EVENT *AllocEvent(MSSONG *pSong,int MaxEvents)
{
	//------------------------------
	// locate a free event
	//
	// parameters
	//	pEv.....pointer to an array of EVENT objects
	//	MaxEvents..number of EVENTS in array
	//
	// return value:
	//	return pointer to free event,
	// or NULL on error (non-free)
	//------------------------------
	int i,loop;
	EVENT *pE=NULL;
	EVENT *pEv = pSong->pEvents;

	for(i=0,loop = 1;(i<MaxEvents) && loop;++i)
	{
		if(pEv[i].EventFree == 0)
		{
			pE =&pEv[i];
			pE->EventFree = 1;	//mark it as IN use
			loop = 0;
			pSong->ActiveEvents ++;	//increment active event count
		}
	}
	return(pE);
}

void FreeEvent(MSSONG *pSong,EVENT *pEv)
{
	pEv->EventFree = 0;
	pSong->ActiveEvents--;	//decrement active event count
}

void do_stop(void)
{
//	unset_timer();
}

int sPrintTrkInfo(char *s,MSSONG *pSong)
{
	int i = 0;
	int j;
	for(j=0;j<15;++j)
		i += sprintf(&s[i],"%2d->%s:#%d OCT=%d OCTn=%d:%s\n",j+1,pSong->trkinfo[j].name,pSong->trkinfo[j].patch,pSong->trkinfo[j].octaveNum,pSong->trkinfo[j].octave,pSong->trkinfo[j].EnableFlag?"ON":"OFF");
	return i;
}

void GetSongParameters(MSSONG *pSong)
{
	int i;
	MIDI_INFO *inf;

	inf = (MIDI_INFO *)&pSong->SongData[MS_OFFSET_MIDI_INFO];
	for(i=0;i<15;++i)
	{
		pSong->trkinfo[i].channel = ((int)((inf->chan_range[i-1]) & 0x0f) - 1) & 0x0f;
		pSong->trkinfo[i].octaveNum = (int)((inf->chan_range[i-1]) >> 4);
		pSong->trkinfo[i].octave = octave[pSong->trkinfo[i].octaveNum];
		pSong->trkinfo[i].patch = (int)(inf->patch[i-1]);
		memcpy(pSong->trkinfo[i].name,inf->Channel[i].Name,10);
	}
	strcpy(pSong->SongName,inf->name);	/*	copy song name	*/
}

void InitBuffer(MSSONG *pSong)
{
	int i;
	EVENT *pEv = pSong->pEvents;

	for(i=0;i<pSong->nEvents;++i,pEv++)
	{
		pEv->Channel = 0;
		pEv->Duration = 0;
		pEv->Note = 0;
		pEv->TieBegFlag = 0;
		pEv->TieEndFlag = 0;
		pEv->EventFree = 0;
		pEv->RestFlag = 0;
		pEv->Velocity = 0;
	}
}

MSSONG * LoadFile(char *SongName,int nEvents)
{
	//------------------------------------
	// load a song file into a song descriptor
	// block.
	// parameters
	//	SongName.....pointer to name of song to load
	//	nEvents.....number of events to allocate
	//--------------------------------------

	struct _finddata_t	FileInfo;
	FILE *pFile;
	long FindHandle;
	int n;
	MSSONG *pSong = NULL;
	long lv,r;

	if((FindHandle = _findfirst(SongName,&FileInfo)) >= 0)
	{
		pFile = fopen(SongName,"rb");
		if(pFile == NULL)
		{
//			printf("Could Not Open %s\n",SongName);
		}
		else
		{
			pSong = (MSSONG *)malloc(sizeof(MSSONG)+FileInfo.size + sizeof(EVENT) * nEvents +8);	//allocate memory for song
			memset(pSong,0,sizeof(MSSONG)+FileInfo.size + sizeof(EVENT) * nEvents + 8);	//zero out memory
			n = fread(pSong->SongData,sizeof(char),FileInfo.size,pFile);
//			printf("File %s[%d] opened, read %d\n",SongName,FileInfo.size,n);
			fclose(pFile);
			_findclose(FindHandle);
			lv = (long)pSong;
			lv += sizeof(MSSONG)+FileInfo.size;
			r = lv % 4;
			lv += r;
			pSong->pEvents = (EVENT *)lv;
			pSong->nEvents = nEvents;
//			printf("Event Buffer Address = %p\n",pSong->pEvents);
		}
	}
	return pSong;
}

//------------------------------
// REPEAT functions
//------------------------------

REPEAT *newRepeat(int cnt, char *pRstrt)
{
	//----------------------------------------
	// This function creates a new
	// REPEAT object
	// parameters:
	//	cnt......number of times to repeat
	//	pRstrt..Start of repeated song data
	// 
	// return value:
	//	returns NULL on failure
	//----------------------------------------
	REPEAT *pR;

	pR = (REPEAT *)malloc(sizeof(REPEAT));
	if(pR)
	{
		pR->Count = cnt;
		pR->pRstart = pRstrt;
	}
	return pR;
}

void deleteRepeat(REPEAT *pR)
{
	free(pR);
}

int DecRepeatCount(REPEAT *pR)
{
	//-------------------------------
	// This function decrements repeat
	// count and then checks to see
	// if it is zero. 
	//
	// paramenters:
	//	pR.....ppointer to repeat object to decrement
	//
	// returns value:
	//	returns false if Count is not zero
	// returns true if count is zero
	//------------------------------
	int rv = 0;

	pR->Count--;
	if(pR->Count == 0)
		rv = 1;
	return rv;
}

//-----------------
// stack functions
//-----------------

STACK *newStack(int size)
{
	//---------------------------------------
	// Create a new stack object
	//
	// parameters
	//	size.....number of elements in stack
	//
	// return value
	//	returns a pointer to the stack object
	// on success.  Returns NULL on fail.
	//---------------------------------------
	STACK *pS;

	pS = (STACK *)malloc(sizeof(STACK) + sizeof(void *) * size);
	if(pS)
	{
		memset(pS,0,sizeof(STACK) + sizeof(void *) * size);
		pS->Size = size;
	}
	return pS;
}

void *GetStackTop(STACK *pS)
{
	//---------------------------------------
	// Gets a pointer to the top object on the
	// stack.
	//
	// parameters
	//	pS.....pointer of stack object to read
	//
	// return value
	//	returns NULL if the stack is empty
	//---------------------------------------
	void *pE = NULL;
	if(pS->nElem)
		pE = pS->data[pS->stactpointer-1];	//get top element
	return pS;
}

void *PopStack(STACK *pS)
{
	//----------------------------------------
	// This function gets the value off the
	// top of the stack and removes it
	//
	// parameters:
	//	pS.....stack to pop object from
	//
	// Return Value
	//	returns NULL if the stack was empty
	//----------------------------------------
	void *pE = NULL;
	if(pS->nElem)
	{
		pE = pS->data[--pS->stactpointer];	//get top element
		pS->nElem--;	//decrement number of elements
	}
	return pS;
}

int PushStack(STACK *pS,void *pE)
{
	//----------------------------------------
	// Push an object onto the stack
	//
	// parameters:
	//	pS.....stack to push object onto
	//	pE....pointer to object to push on stack
	//
	// return value
	//	returns 0 on success, returns 1 if 
	// stack was full
	//-----------------------------------------

	int rv = 0;

	if(pS->nElem == pS->Size)	//stack full?
	{
		rv =1;
	}
	else
	{
		pS->data[pS->stactpointer++] = pE;
		pS->nElem++;	//increment number of elements
	}
	return rv;
}

//-----------------------------------------------
// Initialize function
//-----------------------------------------------

void InitMSplayer(void)
{
	pRepeatStack = newStack(64);	//allow nesting of up to 64 repeats
}

//-----------------------------------------------
// Song Events (SEVENT) functions
//-----------------------------------------------

SEVENT *newSongEvent(int type)
{
	//----------------------------------------
	// this function creates a new SEVENT object
	// 
	// parameters:
	//	type.....type of event
	//	can be the following values
	//			SEVENT_NOTE_TYPE  notes and rests
	//			SEVENT_OBJ_TYPE	  everything else
	//
	// return value:
	//	returns NULL on fail, or a pointer to
	//	an SEVENT object on success
	//----------------------------------------
	SEVENT *pSEv = (SEVENT *)newBucket(sizeof(SEVENT));
	if(pSEv) pSEv->Type = type;
	return pSEv;
}

void PrintSongEvents(MSSONG *pSong)
{
	SEVENT *pSe;
	SEVENT *pTop;
	char *s = (char *)malloc(128);
	pSe = (SEVENT *)GetHead(&pSong->Sevents);
	pSong->MeasureBarCnt =0;
	int EC=0;

	while(pSe)
	{
		pTop = pSe;
		while(pTop)
		{
			switch(pTop->Type)
			{
				case SEVENT_NOTE_TYPE:
					NoteToStr(s,pTop->d.note.Note);
					printf("NOTE=%s  Dur=%s Track = %d TB:%d TE:%d R:%d\n",s,DurTab[pTop->d.note.Duration].pName,pTop->d.note.Track,pTop->d.note.TieBeg,pTop->d.note.TieEnd,pTop->d.note.Rest);
					break;
				case SEVENT_OBJ_TYPE:
					switch(pTop->d.obj.ObjType)
					{
						case SOBJ_TYPE_TIMESIGNATURE:
							printf("Time Signature:%d\n",pTop->d.obj.Data);
							break;
						case SOBJ_TYPE_TEMPO:
							printf("Tempo:%d\n",pTop->d.obj.Data);
							break;
						case SOBJ_TYPE_MEASUREBAR:
							pSong->MeasureBarCnt++;
							printf("Measure Bar %d\n",pSong->MeasureBarCnt);
							break;
						case SOBJ_TYPE_KEY_SIGNATURE:
							printf("KEY Signature:%s\n",KeySigTab[pTop->d.obj.Data]);
							break;
						case SOBJ_TYPE_LOUDNESS:
							printf("Loudness:%d\n",pTop->d.obj.Data);
							break;
						case SOBJ_TYPE_REPEATSTART:
							printf("Repeat START:%d\n",pTop->d.obj.Data);
							break;
						case SOBJ_TYPE_REPEATEND:
							printf("REPEAT end:\n");
							break;
						case SOBJ_TYPE_END:
							printf("*******SONG END********\n");
							break;
					}
					break;
			}
			pTop = pTop->next;
		}
		EC++;
		pSe = (SEVENT *)nextBucket(pSe);
		printf("-------------------%d---------------------\n",EC);
	}
	free(s);
}

int NoteOnOff = 0;

int ProcessEvents(MSSONG *pSong)
{
	EVENT *pEv;
	int i;
	char *s = new char[256];
	int rv = 0;

	pEv = pSong->pEvents;
	for(i=0;i<pSong->nEvents;++i)
	{
		if(pEv->EventFree)
		{
			pEv->Duration--;
			if((pEv->Duration == 2) && !pEv->RestFlag && !pEv->TieBegFlag)
			{
				NoteToStr(s,pEv->Note);
				printf("NOTE OFF %s:%d:EV=%d  Active=%d\n",s,pEv->Track,i,pSong->ActiveEvents);
				NoteOnOff -= 1;;
			}
			else if (pEv->Duration == 0) 
			{
				if(!pEv->TieBegFlag) FreeEvent(pSong,pEv);
				else rv = 1;	//in this case, we need to add more events
			}
			else if(pEv->Duration == -1)	//ERROR?
			{
				printf("ERROR\n");
				pSong->Errors++;
				DisplayEvents(pSong);
				if(!pEv->TieBegFlag)	//may be not such a bad thing
				{
					rv = -1;
					exit(1);
				}
			}
			else if(pEv->Duration < -300)
			{
				//--------------------------
				// if we get here, there is
				// some sort of an error
				// so lets just clean it up
				//--------------------------
				NoteToStr(s,pEv->Note);
				printf("NOTE OFF %s:%d:EV=%d  Active=%d\n",s,pEv->Track,i,pSong->ActiveEvents);
				FreeEvent(pSong,pEv);
				NoteOnOff -= 1;;
			}
		}
		++pEv;
	}
	if(pSong->EndFlag)
	{
		DisplayEvents(pSong);
		printf("NoteOnOff=%d  Active=%d\n",NoteOnOff,pSong->ActiveEvents);
		exit(1);
	}
	delete[]s;
	printf("Exit Process Events RV=%d\n",rv);
	return rv;
}

int AddEvents(MSSONG *pSong,int TickFlag)
{
	SEVENT *pSEv;
	EVENT *pEv;
	int i,printflag =0;
	static char s[128];
	int update =0;
	int rv = 0;
	int Shortest = 288;
	int Longest = 0;

	if(!pSong->EndFlag)
	{
		pSong->Event++;
		printf("--------%d---------\n",pSong->Event++);
		if(pSong->pCurSEv == NULL  )	//have we even started?
		{
			pSong->pCurSEv = (SEVENT *)GetHead(&pSong->Sevents);
		}
		pSEv = pSong->pCurSEv;
		while(pSEv)
		{
			switch(pSEv->Type)
			{
				case SEVENT_NOTE_TYPE:
					if(pSEv->d.note.TieEnd)	//is this note the at the end of a tie?
					{
						pEv = search(pSEv->d.note.Note,pSong->trkinfo[pSEv->d.note.Track].channel,pSong->pEvents,pSong->nEvents);
						if(pEv)
						{
							pEv->TieBegFlag = pSEv->d.note.TieBeg;
							pEv->TieEndFlag = pSEv->d.note.TieEnd;
							pEv->NoteType = pSEv->d.note.Duration;
							pEv->Duration = DurTab[pSEv->d.note.Duration].Dur -1;
							if(pEv->Duration < Shortest)
								Shortest = pEv->Duration;
							if(pEv->Duration > Longest) Longest = pEv->Duration;
							printflag = 1;
							update = 1;
						}
						else
							printf("************* OK BIG ERROR *****************\n");
					}
					else
					{
						pEv = AllocEvent(pSong,pSong->nEvents);	//allocate an event
						pEv->Note = pSEv->d.note.Note;
						pEv->NoteType = pSEv->d.note.Duration;
						pEv->Duration = DurTab[pSEv->d.note.Duration].Dur - 1;
						if(pEv->Duration < Shortest)
							Shortest = pEv->Duration;
						if(pEv->Duration > Longest) Longest = pEv->Duration;
						pEv->Track = pSEv->d.note.Track;
						pEv->Channel = pSong->trkinfo[pSEv->d.note.Track].channel;
						pEv->RestFlag = pSEv->d.note.Rest;
						pEv->TieBegFlag = pSEv->d.note.TieBeg;
						pEv->TieEndFlag = pSEv->d.note.TieEnd;
						pEv->Velocity = pSEv->d.note.Accent?pSong->trkinfo[pEv->Track].velocity + 20:pSong->trkinfo[pEv->Track].velocity;
						if(pEv->Velocity > 127) pEv->Velocity = 127;
						if(!pEv->TieEndFlag && !pEv->RestFlag)
						{
							NoteToStr(s,pEv->Note);
							printf("NOTE ON %s:%d@%d\n",s,pEv->Track,pEv->Velocity);
							NoteOnOff++;
						}
						printflag = 1;
						update = 1;
					}
					break;
				case SEVENT_OBJ_TYPE:
					switch(pSEv->d.obj.ObjType)
					{
						case SOBJ_TYPE_TIMESIGNATURE:
							printf("Time Signature %s\n",TimeSigLut[pSEv->d.obj.Data]);
							break;
						case SOBJ_TYPE_TEMPO:
							pSong->TempoCount = 1200/pSEv->d.obj.Data;
							if(1200 % pSEv->d.obj.Data > pSEv->d.obj.Data /2)
								pSong->TempoCount++;	//round up
							break;
						case SOBJ_TYPE_MEASUREBAR:
							pSong->MeasureBarCnt++;
							printf("Measure Bar %d\n",pSong->MeasureBarCnt);
							break;
						case SOBJ_TYPE_KEY_SIGNATURE:
							printf("Key Sighnature %s\n",KeySigTab[pSEv->d.obj.Data]);
							break;
						case SOBJ_TYPE_LOUDNESS:
							for(i=0;i<15;++i)
							{
								pSong->trkinfo[i].velocity = pSEv->d.obj.Data;
							}
							break;
						case SOBJ_TYPE_REPEATSTART:
							break;
						case SOBJ_TYPE_REPEATEND:
							break;
						case SOBJ_TYPE_END:
							pSong->EndFlag = 1;
							rv = 1;
							break;
					}	//end of switch(pSEv->d.obj.ObjType)
					break;
			}	//end of switch(pSEv->Type)
			pSEv = pSEv->next;
		}//end of while(pSEv)
	}
	else
		printf("------------- END ------------------\n");
	pSong->pCurSEv = (SEVENT *)nextBucket(pSong->pCurSEv);
//	if(TickFlag) pSong->Ticks = Shortest;
	if(TickFlag) pSong->Ticks = Longest;
	return rv;
}

void DisplayEvents(MSSONG *pSong)
{
	EVENT *pEv = pSong->pEvents;
	int i;
	char *s = new char[256];

	printf("xxxxxxxxxxxxxxxxxxxxxxxxxxxx\n");
	if(pSong->ActiveEvents)
	{
		for(i=0;i<pSong->nEvents;++i)
		{
			if(pEv->EventFree)
			{
				NoteToStr(s,pEv->Note);
				printf("%2d::Note=%s::%d  %s Dur=%d TB=%d  TE=%d R=%d\n",i,s,pEv->Channel,DurTab[pEv->NoteType].pName,pEv->Duration,pEv->TieBegFlag,pEv->TieEndFlag,pEv->RestFlag);
			}
			++pEv;
		}
	}
	else
		printf("No Events\n");
	printf("xxxxxxxxxxxxxxxxxxxxxxxxxxxx\n");
	delete  [] s;
}

int Play(MSSONG *pSong)
{
	int printflag =0;
	static char s[128];
	int update =0;
	int rv = 1;
	int ActEv;

	pSong->Ticks--;
	if(ProcessEvents(pSong))
	{
		DisplayEvents(pSong);
		ActEv = pSong->ActiveEvents;
		do
		{
			AddEvents(pSong,pSong->Ticks?0:1);
		} while((pSong->ActiveEvents == ActEv) && (!pSong->EndFlag));
		printf("-------------Event Added When Note Dur -> 0 ----------\n");
		DisplayEvents(pSong);
	}
	else if(pSong->Ticks == 0)
	{
		DisplayEvents(pSong);
		ActEv = pSong->ActiveEvents;
		do
		{
			AddEvents(pSong,1);
		}while((pSong->ActiveEvents == ActEv) && (!pSong->EndFlag));
		printf("-------------Event Added----------\n");
		DisplayEvents(pSong);
		rv = 2;
	}
	if(pSong->EndFlag && (pSong->ActiveEvents == 0) )
	{
		printf("NoteOnOff = %d\n",NoteOnOff);
		rv = 0;
	}
	printf("******** Ticks = %d ********\n",pSong->Ticks);
	if(pSong->Ticks < -1000) exit(1);
	return rv;
}

int ShortestEvent(MSSONG *pSong)
{
	EVENT *pEv = pSong->pEvents;
	int i;
	int rv = 1000;

	for(i=0;i<pSong->nEvents;++i)
	{
		if(pEv[i].EventFree && pEv[i].Duration)
		{
			if(pEv[i].Duration < rv)
				rv = pEv[i].Duration;
		}
	}
	return rv;
}