/************************************************
** Bucket manager
**
** A bucket is just a double linked list/  It
** is possible to attatch any sort of an object
** to a bucket.  When a bucket is created, a
** pointer to the attatched object is returned
**
** Created Feb 21, 20101
**
**************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bucket.h"

void *newBucket(int size)
{
	//-------------------------------------------
	// Thi8s function is used to credate a bucket
	// and to attach another object to it.
	//
	// parameters:
	//	size.....size of the object to attach
	//
	// return value:
	//	returns NULL on fail, else, it returns
	// a pointer to the attatched object
	//-------------------------------------------
	BUCKET *pObj;

	pObj = (BUCKET *)malloc(sizeof(BUCKET) + size);
	if(pObj)
	{
		memset(pObj,0,sizeof(BUCKET) + size);
		pObj++;
	}
	return pObj;

}

void freeBucket(void *pB)
{
	//-------------------------------------------
	// this function destroys a bucket and its
	// attatched objet.
	//
	// pB.....pointer to the attatched object (not the bucket)
	//
	//------------------------------------------
	BUCKET *pObj = (BUCKET *)pB;
	--pObj;
	free(pObj);
}

void delBucket(LIST *pList,void *targ)
{
	//-----------------------------------------
	// remvoes a bucket from a linked list
	//
	// parameters:
	//	pList...list to delecte object from
	//	targ....pointer to object to rememove (not a BUCKET)
	//
	//----------------------------------------
	BUCKET *pD = (BUCKET *)targ;
	pD--;

	if(pD == pList->pTail)
	{
		pList->pTail = pD->pPrev;
		pList->pTail->pNext = NULL;
	}
	else if (pD == pList->pHead)
	{
		pList->pHead = pD->pNext;
		pList->pHead->pPrev = NULL;
	}
	else
	{
		pD->pPrev->pNext = pD->pNext;
		pD->pNext->pPrev = pD->pPrev;
	}
}

void insertBucket(LIST *pList,void *pO,void *pN)
{
	//----------------------------------------
	// Insert an object into a chain.  The
	// new object itself can be a chain of
	// objects
	//
	// parameters:
	//	pList...pointer to list new object to go into
	//	pO.....pointer to object to add new object after
	//	pN.....pointer to new object
	//---------------------------------------

	BUCKET *pOld = (BUCKET *)pO;
	BUCKET *pNew = (BUCKET *)pN;
	BUCKET *tmp;
	pOld--;
	pNew--;
	if(pList->pHead == pOld)
	{
		tmp = pOld->pNext;
		pOld->pNext = pNew;
		pNew->pNext = tmp;
		pNew->pPrev = pOld;
		if(tmp)tmp->pPrev = pNew;
	}
	else if (pList->pTail == pOld)
	{
	}
	else
	{
	}
}

void *nextBucket(void *pO)
{
	//----------------------------------------
	// This function gets the next object in
	// a chian
	// parameter:
	//	pO......pointer to current object
	//
	//	return value:
	//	returns next object on success, or NULL
	//	if there are no more objects
	//---------------------------------------

	BUCKET *pB = (BUCKET *)pO;
	pB--;
	pB = pB->pNext;
	if(pB) ++pB;
	return pB;
}

void *prevBucket(void *pO)
{
	//-----------------------------------------
	// gets the previos object
	//
	// parameters:
	//	pO.....pointer to current object
	//
	// return value
	//	returns prevois object, or NULL if there
	// is no previous object
	//----------------------------------------
	BUCKET *pB = (BUCKET *)pO;
	pB--;
	pB = pB->pPrev;
	if(pB) ++pB;
	return pB;
}

void AddObject(LIST *pList,void *pNew)
{
	//----------------------------------------------
	// Add a new object to the end of the chain
	//
	// parameters:
	//	pTail....pointer to the end of the chain
	//	pNew....new object to add to the end of the chain
	//----------------------------------------------
	BUCKET *tmp;
	BUCKET *pO = (BUCKET *)pNew;

	pO--;
	tmp = pList->pTail;
	pList->pTail = pO;
	pO->pNext = 0;
	pO->pPrev = tmp;
	if(tmp)
		tmp->pNext = pO;
	if(pList->pHead == NULL)
		pList->pHead = pList->pTail;
}

void *GetHead(LIST *pList)
{
	return (void *)(pList->pHead + 1);
}

void *GetTail(LIST *pList)
{
	return (void *)(pList->pTail + 1);
}
